# -*- coding: utf-8 -*-
from __future__ import absolute_import

from .reader import Reader
from .nsqd import Nsqd
from .lookupd import Lookupd
from .message import Message
from .backofftimer import BackoffTimer
from .version import __version__

__author__ = 'Trevor Olson'
__email__ = 'trevor@heytrevor.com'
__version__ = __version__

__all__ = [
    'Reader',
    'Nsqd',
    'Lookupd',
    'Message',
    'BackoffTimer',
]



from __future__ import with_statement

import pytest
import gevent
import gnsq

from integration_server import (
    with_all,
    LookupdIntegrationServer,
    NsqdIntegrationServer
)


@pytest.mark.slow
def test_basic():
    with LookupdIntegrationServer() as server:
        lookupd = gnsq.Lookupd(server.http_address)
        assert lookupd.ping() == 'OK'
        assert 'version' in lookupd.info()

        with pytest.raises(gnsq.errors.NSQHttpError):
            lookupd.lookup('topic')

        assert len(lookupd.topics()['topics']) == 0
        assert len(lookupd.channels('topic')['channels']) == 0
        assert len(lookupd.nodes()['producers']) == 0


@pytest.mark.slow
def test_lookup():
    lookupd_server = LookupdIntegrationServer()
    nsqd_server = NsqdIntegrationServer(lookupd=lookupd_server.tcp_address)

    @with_all(lookupd_server, nsqd_server)
    def _(lookupd_server, nsqd_server):
        lookupd = gnsq.Lookupd(lookupd_server.http_address)
        conn = gnsq.Nsqd(nsqd_server.address, http_port=nsqd_server.http_port)

        assert len(lookupd.topics()['topics']) == 0
        assert len(lookupd.channels('topic')['channels']) == 0
        assert len(lookupd.nodes()['producers']) == 1

        conn.create_topic('topic')
        gevent.sleep(0.1)

        info = lookupd.lookup('topic')
        assert len(info['channels']) == 0
        assert len(info['producers']) == 1
        assert len(lookupd.topics()['topics']) == 1
        assert len(lookupd.channels('topic')['channels']) == 0

        conn.create_channel('topic', 'channel')
        gevent.sleep(0.1)

        info = lookupd.lookup('topic')
        assert len(info['channels']) == 1
        assert len(info['producers']) == 1
        assert len(lookupd.topics()['topics']) == 1
        assert len(lookupd.channels('topic')['channels']) == 1



import pytest


def pytest_addoption(parser):
    parser.addoption(
        '--fast',
        action='store_true',
        help='do not run slow tests'
    )


def pytest_runtest_setup(item):
    if 'slow' in item.keywords and item.config.getoption('--fast'):
        pytest.skip('skiping because of --fast')



from __future__ import with_statement

import multiprocessing
import pytest
import gevent

from gnsq import Nsqd, Reader, states
from gnsq.errors import NSQSocketError

from integration_server import (
    with_all,
    LookupdIntegrationServer,
    NsqdIntegrationServer
)


def test_basic():
    with pytest.raises(ValueError):
        Reader('test', 'test')

    with pytest.raises(TypeError):
        Reader(
            topic='test',
            channel='test',
            nsqd_tcp_addresses=None,
            lookupd_http_addresses='http://localhost:4161/',
        )

    with pytest.raises(TypeError):
        Reader(
            topic='test',
            channel='test',
            nsqd_tcp_addresses='localhost:4150',
            lookupd_http_addresses=None,
        )

    def message_handler(reader, message):
        pass

    reader = Reader(
        topic='test',
        channel='test',
        name='test',
        max_concurrency=-1,
        nsqd_tcp_addresses='localhost:4150',
        lookupd_http_addresses='http://localhost:4161/',
        message_handler=message_handler
    )

    assert reader.name == 'test'
    assert reader.max_concurrency == multiprocessing.cpu_count()
    assert len(reader.on_message.receivers) == 1

    assert isinstance(reader.nsqd_tcp_addresses, set)
    assert len(reader.nsqd_tcp_addresses) == 1

    assert isinstance(reader.lookupds, list)
    assert len(reader.lookupds) == 1


@pytest.mark.slow
def test_messages():
    with NsqdIntegrationServer() as server:

        class Accounting(object):
            count = 0
            total = 500
            error = None

        conn = Nsqd(
            address=server.address,
            tcp_port=server.tcp_port,
            http_port=server.http_port,
        )

        for _ in xrange(Accounting.total):
            conn.publish_http('test', 'danger zone!')

        reader = Reader(
            topic='test',
            channel='test',
            nsqd_tcp_addresses=[server.tcp_address],
            max_in_flight=100,
        )

        @reader.on_exception.connect
        def error_handler(reader, message, error):
            if isinstance(error, NSQSocketError):
                return
            Accounting.error = error
            reader.close()

        @reader.on_message.connect
        def handler(reader, message):
            assert message.body == 'danger zone!'

            Accounting.count += 1
            if Accounting.count == Accounting.total:
                assert not reader.is_starved
                reader.close()

        try:
            reader.start()
        except NSQSocketError:
            pass

        if Accounting.error:
            raise Accounting.error

        assert Accounting.count == Accounting.total


@pytest.mark.slow
def test_max_concurrency():
    server1 = NsqdIntegrationServer()
    server2 = NsqdIntegrationServer()

    @with_all(server1, server2)
    def _(server1, server2):
        class Accounting(object):
            count = 0
            total = 100
            concurrency = 0
            error = None

        for server in (server1, server2):
            conn = Nsqd(
                address=server.address,
                tcp_port=server.tcp_port,
                http_port=server.http_port,
            )

            for _ in xrange(Accounting.total / 2):
                conn.publish_http('test', 'danger zone!')

        reader = Reader(
            topic='test',
            channel='test',
            nsqd_tcp_addresses=[
                server1.tcp_address,
                server2.tcp_address,
            ],
            max_in_flight=5,
            max_concurrency=1,
        )

        @reader.on_exception.connect
        def error_handler(reader, message, error):
            if isinstance(error, NSQSocketError):
                return
            Accounting.error = error
            reader.close()

        @reader.on_message.connect
        def handler(reader, message):
            assert message.body == 'danger zone!'
            assert Accounting.concurrency == 0

            Accounting.concurrency += 1
            gevent.sleep()
            Accounting.concurrency -= 1

            Accounting.count += 1
            if Accounting.count == Accounting.total:
                reader.close()

        try:
            reader.start()
        except NSQSocketError:
            pass

        if Accounting.error:
            raise Accounting.error

        assert Accounting.count == Accounting.total


@pytest.mark.slow
def test_lookupd():
    lookupd_server = LookupdIntegrationServer()
    server1 = NsqdIntegrationServer(lookupd=lookupd_server.tcp_address)
    server2 = NsqdIntegrationServer(lookupd=lookupd_server.tcp_address)

    @with_all(lookupd_server, server1, server2)
    def _(lookupd_server, server1, server2):
        class Accounting(object):
            count = 0
            total = 500
            concurrency = 0
            error = None

        for server in (server1, server2):
            conn = Nsqd(
                address=server.address,
                tcp_port=server.tcp_port,
                http_port=server.http_port,
            )

            for _ in xrange(Accounting.total / 2):
                conn.publish_http('test', 'danger zone!')

        reader = Reader(
            topic='test',
            channel='test',
            lookupd_http_addresses=lookupd_server.http_address,
            max_in_flight=32,
        )

        @reader.on_exception.connect
        def error_handler(reader, message, error):
            if isinstance(error, NSQSocketError):
                return
            Accounting.error = error
            reader.close()

        @reader.on_message.connect
        def handler(reader, message):
            assert message.body == 'danger zone!'

            Accounting.count += 1
            if Accounting.count == Accounting.total:
                reader.close()

        try:
            reader.start()
        except NSQSocketError:
            pass

        if Accounting.error:
            raise Accounting.error

        assert Accounting.count == Accounting.total


@pytest.mark.slow
def test_backoff():
    with NsqdIntegrationServer() as server:
        conn = Nsqd(
            address=server.address,
            tcp_port=server.tcp_port,
            http_port=server.http_port,
        )

        for _ in xrange(500):
            conn.publish_http('test', 'danger zone!')

        reader = Reader(
            topic='test',
            channel='test',
            nsqd_tcp_addresses=[server.tcp_address],
            max_in_flight=100,
            message_handler=lambda reader, message: None
        )

        reader.start(block=False)
        reader.start_backoff()

        assert reader.state == states.THROTTLED
        assert reader.total_in_flight_or_ready <= 1

        reader.complete_backoff()
        assert reader.state == states.RUNNING


def test_no_handlers():
    reader = Reader('test', 'test', 'localhost:4150')
    with pytest.raises(RuntimeError):
        reader.start(block=False)



from __future__ import with_statement

import struct
import json
import ssl
import pytest

from itertools import product
from gnsq import Nsqd, Message, states, errors
from gnsq import protocol as nsq
from gnsq.stream.stream import SSLSocket, DefalteSocket, SnappySocket

from mock_server import mock_server
from integration_server import NsqdIntegrationServer


def mock_response(frame_type, data):
    body_size = 4 + len(data)
    body_size_packed = struct.pack('>l', body_size)
    frame_type_packed = struct.pack('>l', frame_type)
    return body_size_packed + frame_type_packed + data


def mock_response_message(timestamp, attempts, id, body):
    timestamp_packed = struct.pack('>q', timestamp)
    attempts_packed = struct.pack('>h', attempts)
    id = "%016d" % id
    data = timestamp_packed + attempts_packed + id + body
    return mock_response(nsq.FRAME_TYPE_MESSAGE, data)


def test_connection():
    @mock_server
    def handle(socket, address):
        assert socket.recv(4) == '  V2'
        assert socket.recv(1) == ''

    with handle as server:
        conn = Nsqd(address='127.0.0.1', tcp_port=server.server_port)
        assert conn.state == states.INIT

        conn.connect()
        assert conn.state == states.CONNECTED

        conn.connect()
        assert conn.state == states.CONNECTED

        conn.close_stream()
        assert conn.state == states.DISCONNECTED


def test_disconnected():
    @mock_server
    def handle(socket, address):
        assert socket.recv(4) == '  V2'
        assert socket.recv(1) == ''

    with handle as server:
        conn = Nsqd(address='127.0.0.1', tcp_port=server.server_port)
        conn.connect()
        conn.close_stream()
        assert conn.state == states.DISCONNECTED

        with pytest.raises(errors.NSQSocketError):
            conn.nop()

        with pytest.raises(errors.NSQSocketError):
            conn.read_response()


@pytest.mark.parametrize('body', [
    'hello world',
    '',
    '{"some": "json data"}',
])
def test_read(body):
    @mock_server
    def handle(socket, address):
        socket.send(struct.pack('>l', len(body)))
        socket.send(body)

    with handle as server:
        conn = Nsqd(address='127.0.0.1', tcp_port=server.server_port)
        conn.connect()

        assert conn._read_response() == body
        conn.close_stream()


def test_identify():
    @mock_server
    def handle(socket, address):
        assert socket.recv(4) == '  V2'
        assert socket.recv(9) == 'IDENTIFY\n'

        size = nsq.unpack_size(socket.recv(4))
        data = json.loads(socket.recv(size))

        assert 'gnsq' in data['user_agent']
        socket.send(mock_response(nsq.FRAME_TYPE_RESPONSE, 'OK'))

    with handle as server:
        conn = Nsqd(address='127.0.0.1', tcp_port=server.server_port)
        conn.connect()

        assert conn.identify() is None


def test_negotiation():
    @mock_server
    def handle(socket, address):
        assert socket.recv(4) == '  V2'
        assert socket.recv(9) == 'IDENTIFY\n'

        size = nsq.unpack_size(socket.recv(4))
        data = json.loads(socket.recv(size))

        assert 'gnsq' in data['user_agent']
        resp = json.dumps({'test': 42})
        socket.send(mock_response(nsq.FRAME_TYPE_RESPONSE, resp))

    with handle as server:
        conn = Nsqd(address='127.0.0.1', tcp_port=server.server_port)
        conn.connect()

        assert conn.identify()['test'] == 42


@pytest.mark.parametrize('command,args,resp', [
    ('subscribe', ('topic', 'channel'), 'SUB topic channel\n'),
    ('subscribe', ('foo', 'bar'), 'SUB foo bar\n'),
    ('ready', (0,), 'RDY 0\n'),
    ('ready', (1,), 'RDY 1\n'),
    ('ready', (42,), 'RDY 42\n'),
    ('finish', ('0000000000000000',), 'FIN 0000000000000000\n'),
    ('finish', ('deadbeafdeadbeaf',), 'FIN deadbeafdeadbeaf\n'),
    ('requeue', ('0000000000000000',), 'REQ 0000000000000000 0\n'),
    ('requeue', ('deadbeafdeadbeaf', 0), 'REQ deadbeafdeadbeaf 0\n'),
    ('requeue', ('deadbeafdeadbeaf', 42), 'REQ deadbeafdeadbeaf 42\n'),
    ('touch', ('0000000000000000',), 'TOUCH 0000000000000000\n'),
    ('touch', ('deadbeafdeadbeaf',), 'TOUCH deadbeafdeadbeaf\n'),
    ('close', (), 'CLS\n'),
    ('nop', (), 'NOP\n'),
])
def test_command(command, args, resp):
    @mock_server
    def handle(socket, address):
        assert socket.recv(4) == '  V2'
        assert socket.recv(len(resp)) == resp

    with handle as server:
        conn = Nsqd(address='127.0.0.1', tcp_port=server.server_port)
        conn.connect()
        getattr(conn, command)(*args)


def test_publish():
    @mock_server
    def handle(socket, address):
        assert socket.recv(4) == '  V2'
        assert socket.recv(10) == 'PUB topic\n'

        assert nsq.unpack_size(socket.recv(4)) == 3
        assert socket.recv(3) == 'sup'

    with handle as server:
        conn = Nsqd(address='127.0.0.1', tcp_port=server.server_port)
        conn.connect()
        conn.publish('topic', 'sup')


def test_multipublish():
    @mock_server
    def handle(socket, address):
        assert socket.recv(4) == '  V2'
        assert socket.recv(11) == 'MPUB topic\n'

        size = nsq.unpack_size(socket.recv(4))
        data = socket.recv(size)

        head, data = data[:4], data[4:]
        assert nsq.unpack_size(head) == 2

        for _ in xrange(2):
            head, data = data[:4], data[4:]
            assert nsq.unpack_size(head) == 3

            head, data = data[:3], data[3:]
            assert head == 'sup'

        assert data == ''

    with handle as server:
        conn = Nsqd(address='127.0.0.1', tcp_port=server.server_port)
        conn.connect()
        conn.multipublish('topic', ['sup', 'sup'])


@pytest.mark.parametrize('error_msg,error,fatal', [
    ('E_INVALID cannot SUB in current state', 'NSQInvalid', True),
    ('E_BAD_BODY MPUB failed to read body size', 'NSQBadBody', True),
    ('E_BAD_TOPIC SUB topic name oh my god is not valid', 'NSQBadTopic', True),
    ('E_BAD_CHANNEL SUB channel name !! is not valid', 'NSQBadChannel', True),
    ('E_BAD_MESSAGE PUB failed to read message body', 'NSQBadMessage', True),
    ('E_PUT_FAILED PUT failed', 'NSQPutFailed', True),
    ('E_PUB_FAILED PUB failed', 'NSQPubFailed', True),
    ('E_MPUB_FAILED MPUB failed', 'NSQMPubFailed', True),
    ('E_AUTH_DISABLED AUTH Disabled', 'NSQAuthDisabled', True),
    ('E_AUTH_FAILED AUTH failed', 'NSQAuthFailed', True),
    ('E_UNAUTHORIZED AUTH No authorizations found', 'NSQUnauthorized', True),
    ('E_FIN_FAILED FIN failed', 'NSQFinishFailed', False),
    ('E_REQ_FAILED REQ failed', 'NSQRequeueFailed', False),
    ('E_TOUCH_FAILED TOUCH failed', 'NSQTouchFailed', False),
    ('some unknown error', 'NSQException', True),
])
def test_error(error_msg, error, fatal):
    @mock_server
    def handle(socket, address):
        assert socket.recv(4) == '  V2'
        socket.send(mock_response(nsq.FRAME_TYPE_ERROR, error_msg))

    with handle as server:
        conn = Nsqd(address='127.0.0.1', tcp_port=server.server_port)
        conn.connect()

        frame, resp = conn.read_response()
        assert frame == nsq.FRAME_TYPE_ERROR
        assert isinstance(resp, getattr(errors, error))
        assert conn.is_connected != fatal


def test_hashing():
    conn1 = Nsqd('localhost', 1337)
    conn2 = Nsqd('localhost', 1337)
    assert conn1 == conn2
    assert not (conn1 < conn2)
    assert not (conn2 < conn1)

    test = {conn1: True}
    assert conn2 in test


def test_sync_receive_messages():
    @mock_server
    def handle(socket, address):
        assert socket.recv(4) == '  V2'
        assert socket.recv(9) == 'IDENTIFY\n'

        size = nsq.unpack_size(socket.recv(4))
        data = json.loads(socket.recv(size))

        assert isinstance(data, dict)
        socket.send(mock_response(nsq.FRAME_TYPE_RESPONSE, 'OK'))

        msg = 'SUB topic channel\n'
        assert socket.recv(len(msg)) == msg
        socket.send(mock_response(nsq.FRAME_TYPE_RESPONSE, 'OK'))

        for i in xrange(10):
            assert socket.recv(6) == 'RDY 1\n'

            body = json.dumps({'data': {'test_key': i}})
            ts = i * 1000 * 1000
            socket.send(mock_response_message(ts, i, i, body))

    with handle as server:
        conn = Nsqd(address='127.0.0.1', tcp_port=server.server_port)
        conn.connect()

        assert conn.identify() is None

        conn.subscribe('topic', 'channel')
        frame, data = conn.read_response()

        assert frame == nsq.FRAME_TYPE_RESPONSE
        assert data == 'OK'

        for i in xrange(10):
            conn.ready(1)
            frame, msg = conn.read_response()

            assert frame == nsq.FRAME_TYPE_MESSAGE
            assert isinstance(msg, Message)
            assert msg.timestamp == i * 1000 * 1000
            assert msg.id == '%016d' % i
            assert msg.attempts == i
            assert json.loads(msg.body)['data']['test_key'] == i


def test_sync_heartbeat():
    @mock_server
    def handle(socket, address):
        assert socket.recv(4) == '  V2'
        socket.send(mock_response(nsq.FRAME_TYPE_RESPONSE, '_heartbeat_'))
        assert socket.recv(4) == 'NOP\n'

    with handle as server:
        conn = Nsqd(address='127.0.0.1', tcp_port=server.server_port)
        conn.connect()

        frame, data = conn.read_response()
        assert frame == nsq.FRAME_TYPE_RESPONSE
        assert data == '_heartbeat_'


def test_auth():
    @mock_server
    def handle(socket, address):
        assert socket.recv(4) == '  V2'
        assert socket.recv(5) == 'AUTH\n'

        assert nsq.unpack_size(socket.recv(4)) == 6
        assert socket.recv(6) == 'secret'

        resp = json.dumps({'identity': 'awesome'})
        socket.send(mock_response(nsq.FRAME_TYPE_RESPONSE, resp))

    with handle as server:
        conn = Nsqd(
            address='127.0.0.1',
            tcp_port=server.server_port,
            auth_secret='secret'
        )

        conn.connect()
        resp = conn.auth()
        assert resp['identity'] == 'awesome'


def test_identify_auth():
    @mock_server
    def handle(socket, address):
        assert socket.recv(4) == '  V2'
        assert socket.recv(9) == 'IDENTIFY\n'

        size = nsq.unpack_size(socket.recv(4))
        data = json.loads(socket.recv(size))
        assert 'gnsq' in data['user_agent']

        resp = json.dumps({'auth_required': True})
        socket.send(mock_response(nsq.FRAME_TYPE_RESPONSE, resp))

        assert socket.recv(5) == 'AUTH\n'
        assert nsq.unpack_size(socket.recv(4)) == 6
        assert socket.recv(6) == 'secret'

        resp = json.dumps({'identity': 'awesome'})
        socket.send(mock_response(nsq.FRAME_TYPE_RESPONSE, resp))

    with handle as server:
        conn = Nsqd(
            address='127.0.0.1',
            tcp_port=server.server_port,
            auth_secret='secret'
        )

        @conn.on_auth.connect
        def assert_auth(conn, response):
            assert assert_auth.was_called is False
            assert_auth.was_called = True
            assert response['identity'] == 'awesome'

        assert_auth.was_called = False
        conn.connect()
        resp = conn.identify()

        assert resp['auth_required']
        assert assert_auth.was_called


@pytest.mark.parametrize('tls,deflate,snappy', product((True, False), repeat=3))
@pytest.mark.slow
def test_socket_upgrades(tls, deflate, snappy):
    with NsqdIntegrationServer() as server:
        options = {
            'address': server.address,
            'tcp_port': server.tcp_port,
            'deflate': deflate,
            'snappy': snappy,
        }

        if tls:
            options.update({
                'tls_v1': True,
                'tls_options': {
                    'keyfile': server.tls_key,
                    'certfile': server.tls_cert,
                }
            })

        conn = Nsqd(**options)
        conn.connect()
        assert conn.state == states.CONNECTED

        if deflate and snappy:
            with pytest.raises(errors.NSQErrorCode):
                conn.identify()
            return

        if tls and server.version < (0, 2, 28):
            with pytest.raises(ssl.SSLError):
                conn.identify()
            return

        resp = conn.identify()
        assert isinstance(resp, dict)

        assert resp['tls_v1'] is tls
        assert resp['deflate'] is deflate
        assert resp['snappy'] is snappy

        if tls and (deflate or snappy):
            assert isinstance(conn.stream.socket._socket, SSLSocket)
        elif tls:
            assert isinstance(conn.stream.socket, SSLSocket)

        if deflate:
            assert isinstance(conn.stream.socket, DefalteSocket)

        if snappy:
            assert isinstance(conn.stream.socket, SnappySocket)

        conn.publish('topic', 'sup')
        frame, data = conn.read_response()
        assert frame == nsq.FRAME_TYPE_RESPONSE
        assert data == 'OK'

        conn.subscribe('topic', 'channel')
        frame, data = conn.read_response()
        assert frame == nsq.FRAME_TYPE_RESPONSE
        assert data == 'OK'

        conn.ready(1)
        frame, data = conn.read_response()
        assert frame == nsq.FRAME_TYPE_MESSAGE
        assert data.body == 'sup'

        conn.close_stream()


@pytest.mark.slow
def test_cls_error():
    with NsqdIntegrationServer() as server:
        conn = Nsqd(address=server.address, tcp_port=server.tcp_port)

        conn.connect()
        assert conn.state == states.CONNECTED

        conn.close()
        frame, error = conn.read_response()
        assert frame == nsq.FRAME_TYPE_ERROR
        assert isinstance(error, errors.NSQInvalid)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from resource import getpagesize
from errno import ENOTCONN, EDEADLK, EAGAIN, EWOULDBLOCK

import gevent
from gevent import socket
from gevent.queue import Queue
from gevent.event import AsyncResult

try:
    from gevent.ssl import SSLSocket, PROTOCOL_TLSv1, CERT_NONE
except ImportError:
    SSLSocket = None  # pyflakes.ignore

from gnsq.states import INIT, CONNECTED, DISCONNECTED
from gnsq.errors import NSQSocketError

try:
    from .snappy import SnappySocket
except ImportError:
    SnappySocket = None  # pyflakes.ignore

from .defalte import DefalteSocket


class Stream(object):
    def __init__(self, address, port, timeout, buffer_size=getpagesize()):
        self.address = address
        self.port = port
        self.timeout = timeout

        self.buffer = ''
        self.buffer_size = buffer_size

        self.socket = None
        self.worker = None
        self.queue = Queue()
        self.state = INIT

    @property
    def is_connected(self):
        return self.state == CONNECTED

    def ensure_connection(self):
        if self.is_connected:
            return
        raise NSQSocketError(ENOTCONN, 'Socket is not connected')

    def connect(self):
        if self.state not in (INIT, DISCONNECTED):
            return

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(self.timeout)

        try:
            self.socket.connect((self.address, self.port))
        except socket.error as error:
            raise NSQSocketError(*error)

        self.state = CONNECTED
        self.worker = gevent.spawn(self.send_loop)

    def read(self, size):
        while len(self.buffer) < size:
            self.ensure_connection()

            try:
                packet = self.socket.recv(self.buffer_size)
            except socket.error as error:
                if error.errno in (EDEADLK, EAGAIN, EWOULDBLOCK):
                    gevent.sleep()
                    continue
                raise NSQSocketError(*error)

            if not packet:
                self.close()

            self.buffer += packet

        data = self.buffer[:size]
        self.buffer = self.buffer[size:]

        return data

    def send(self, data, async=False):
        self.ensure_connection()

        result = AsyncResult()
        self.queue.put((data, result))

        if async:
            return result

        result.get()

    def consume_buffer(self):
        data = self.buffer
        self.buffer = ''
        return data

    def close(self):
        if not self.is_connected:
            return

        self.state = DISCONNECTED
        self.queue.put(StopIteration)

    def send_loop(self):
        for data, result in self.queue:
            if not self.is_connected:
                error = NSQSocketError(ENOTCONN, 'Socket is not connected')
                result.set_exception(error)

            try:
                self.socket.send(data)
                result.set()

            except socket.error as error:
                result.set_exception(NSQSocketError(*error))

            except Exception as error:
                result.set_exception(error)

        self.socket.close()

    def upgrade_to_tls(
        self,
        keyfile=None,
        certfile=None,
        cert_reqs=CERT_NONE,
        ca_certs=None
    ):
        if SSLSocket is None:
            msg = 'tls_v1 requires Python 2.6+ or Python 2.5 w/ pip install ssl'
            raise RuntimeError(msg)

        self.ensure_connection()
        self.socket = SSLSocket(
            self.socket,
            keyfile=keyfile,
            certfile=certfile,
            cert_reqs=cert_reqs,
            ca_certs=ca_certs,
            ssl_version=PROTOCOL_TLSv1,
        )

    def upgrade_to_snappy(self):
        if SnappySocket is None:
            raise RuntimeError('snappy requires the python-snappy package')

        self.ensure_connection()
        self.socket = SnappySocket(self.socket)
        self.socket.bootstrap(self.consume_buffer())

    def upgrade_to_defalte(self, level):
        self.ensure_connection()
        self.socket = DefalteSocket(self.socket, level)
        self.socket.bootstrap(self.consume_buffer())



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import urllib3

try:
    import simplejson as json
except ImportError:
    import json  # pyflakes.ignore

from .decorators import cached_property
from .errors import NSQHttpError


class HTTPClient(object):
    @cached_property
    def http(self):
        return urllib3.connection_from_url(url=self.base_url)

    def http_request(self, method, url, **kwargs):
        response = self.http.request_encode_url(method, url, **kwargs)

        if 'application/json' in response.getheader('content-type', ''):
            return self._http_check_json(response)

        return self._http_check(response)

    def _http_check(self, response):
        if response.status != 200:
            raise NSQHttpError('http error <%s>' % response.status)
        return response.data

    def _http_check_json(self, response):
        try:
            data = json.loads(response.data)
        except ValueError:
            return self._http_check(response)

        if response.status != 200:
            status_txt = data.get('status_txt', 'http error')
            raise NSQHttpError('%s <%s>' % (status_txt, response.status))

        return data['data']

    def http_get(self, url, **kwargs):
        return self.http_request('GET', url, **kwargs)

    def http_post(self, url, **kwargs):
        return self.http_request('POST', url, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging
import random
import gevent
import blinker
import time

from itertools import cycle
from collections import defaultdict
from multiprocessing import cpu_count
from gevent.queue import Queue

from .lookupd import Lookupd
from .nsqd import Nsqd
from .backofftimer import BackoffTimer
from .states import INIT, RUNNING, BACKOFF, THROTTLED, CLOSED
from .decorators import cached_property

from .errors import (
    NSQException,
    NSQNoConnections,
    NSQRequeueMessage
)


class Reader(object):
    """High level NSQ consumer.

    A Reader will connect to the nsqd tcp addresses or poll the provided
    nsqlookupd http addresses for the configured topic and send signals to
    message handlers connected to the `on_message` signal or provided by
    `message_handler`.

    Messages will automatically be finished when the message handle returns
    unless the readers `async` flag is set to `True`. If an exception occurs or
    :class:`gnsq.errors.NSQRequeueMessage` is raised, the message will be
    requeued.

    The Reader will handle backing off of failed messages up to a configurable
    `max_interval` as well as automatically reconnecting to dropped connections.

    :param topic: specifies the desired NSQ topic

    :param channel: specifies the desired NSQ channel

    :param nsqd_tcp_addresses: a sequence of string addresses of the nsqd
        instances this reader should connect to

    :param lookupd_http_addresses: a sequence of string addresses of the
        nsqlookupd instances this reader should query for producers of the
        specified topic

    :param name: a string that is used for logging messages (defaults to
        'gnsq.reader.topic.channel')

    :param message_handler: the callable that will be executed for each message
        received

    :param async: consider the message handling to be async. The message will
        not automatically be finished after the handler returns and must
        manually be called

    :param max_tries: the maximum number of attempts the reader will make to
        process a message after which messages will be automatically discarded

    :param max_in_flight: the maximum number of messages this reader will
        pipeline for processing. this value will be divided evenly amongst the
        configured/discovered nsqd producers

    :param max_concurrency: the maximum number of messages that will be handled
        concurrently. Defaults to the number of nsqd connections. Setting
        `max_concurrency` to `-1` will use the systems cpu count.

    :param requeue_delay: the default delay to use when requeueing a failed
        message

    :param lookupd_poll_interval: the amount of time in seconds between querying
        all of the supplied nsqlookupd instances.  A random amount of time based
        on this value will be initially introduced in order to add jitter when
        multiple readers are running

    :param lookupd_poll_jitter: the maximum fractional amount of jitter to add
        to the lookupd pool loop. This helps evenly distribute requests even if
        multiple consumers restart at the same time.

    :param low_ready_idle_timeout: the amount of time in seconds to wait for a
        message from a producer when in a state where RDY counts are
        re-distributed (ie. max_in_flight < num_producers)

    :param max_backoff_duration: the maximum time we will allow a backoff state
        to last in seconds

    :param \*\*kwargs: passed to :class:`gnsq.Nsqd` initialization
    """
    def __init__(
        self,
        topic,
        channel,
        nsqd_tcp_addresses=[],
        lookupd_http_addresses=[],
        name=None,
        message_handler=None,
        async=False,
        max_tries=5,
        max_in_flight=1,
        max_concurrency=0,
        requeue_delay=0,
        lookupd_poll_interval=60,
        lookupd_poll_jitter=0.3,
        low_ready_idle_timeout=10,
        max_backoff_duration=128,
        **kwargs
    ):
        if not nsqd_tcp_addresses and not lookupd_http_addresses:
            raise ValueError('must specify at least on nsqd or lookupd')

        nsqd_tcp_addresses = self._get_nsqds(nsqd_tcp_addresses)
        lookupd_http_addresses = self._get_lookupds(lookupd_http_addresses)
        random.shuffle(lookupd_http_addresses)

        self.nsqd_tcp_addresses = nsqd_tcp_addresses
        self.lookupds = [Lookupd(a) for a in lookupd_http_addresses]
        self.iterlookupds = cycle(self.lookupds)

        self.topic = topic
        self.channel = channel
        self.async = async
        self.max_tries = max_tries
        self.max_in_flight = max_in_flight
        self.requeue_delay = requeue_delay
        self.lookupd_poll_interval = lookupd_poll_interval
        self.lookupd_poll_jitter = lookupd_poll_jitter
        self.low_ready_idle_timeout = low_ready_idle_timeout
        self.max_backoff_duration = max_backoff_duration
        self.conn_kwargs = kwargs

        if name:
            self.name = name
        else:
            self.name = '%s.%s.%s' % (__name__, self.topic, self.channel)

        if message_handler is not None:
            self.on_message.connect(message_handler, weak=False)

        if max_concurrency < 0:
            self.max_concurrency = cpu_count()
        else:
            self.max_concurrency = max_concurrency

        if max_concurrency:
            self.queue = Queue()
        else:
            self.queue = None

        self._need_ready_redistributed = False
        self.last_random_ready = time.time()
        self.state = INIT

        self.logger = logging.getLogger(self.name)
        self.conn_backoffs = defaultdict(self.create_backoff)
        self.backoff = self.create_backoff()

        self.conns = set()
        self.pending = set()

        self.workers = []
        self.conn_workers = {}

    @cached_property
    def on_message(self):
        """Emitted when a message is received.

        The signal sender is the reader and the `message` is sent as an
        argument. The `message_handler` param is connected to this signal.
        """
        return blinker.Signal(doc='Emitted when a message is received.')

    @cached_property
    def on_response(self):
        """Emitted when a response is received.

        The signal sender is the reader and the `response` is sent as an
        argument.
        """
        return blinker.Signal(doc='Emitted when a response is received.')

    @cached_property
    def on_error(self):
        """Emitted when an error is received.

        The signal sender is the reader and the `error` is sent as an
        argument.
        """
        return blinker.Signal(doc='Emitted when a error is received.')

    @cached_property
    def on_finish(self):
        """Emitted after a message is successfully finished.

        The signal sender is the reader and the `message_id` is sent as an
        argument.
        """
        return blinker.Signal(doc='Emitted after the a message is finished.')

    @cached_property
    def on_requeue(self):
        """Emitted after a message is requeued.

        The signal sender is the reader and the `message_id` and `timeout`
        are sent as arguments.
        """
        return blinker.Signal(doc='Emitted after the a message is requeued.')

    @cached_property
    def on_giving_up(self):
        """Emitted after a giving up on a message.

        Emitted when a message has exceeded the maximum number of attempts
        (`max_tries`) and will no longer be requeued. This is useful to perform
        tasks such as writing to disk, collecting statistics etc. The signal
        sender is the reader and the `message` is sent as an argument.
        """
        return blinker.Signal(doc='Sent after a giving up on a message.')

    @cached_property
    def on_auth(self):
        """Emitted after a connection is successfully authenticated.

        The signal sender is the reader and the `conn` and parsed `response` are
        sent as arguments.
        """
        return blinker.Signal(doc='Emitted when a response is received.')

    @cached_property
    def on_exception(self):
        """Emitted when an exception is caught while handling a message.

        The signal sender is the reader and the `message` and `error` are sent
        as arguments.
        """
        return blinker.Signal(doc='Emitted when an exception is caught.')

    @cached_property
    def on_close(self):
        """Emitted after :meth:`close`.

        The signal sender is the reader.
        """
        return blinker.Signal(doc='Emitted after the reader is closed.')

    def _get_nsqds(self, nsqd_tcp_addresses):
        if isinstance(nsqd_tcp_addresses, basestring):
            return set([nsqd_tcp_addresses])

        elif isinstance(nsqd_tcp_addresses, (list, tuple, set)):
            return set(nsqd_tcp_addresses)

        raise TypeError('nsqd_tcp_addresses must be a list, set or tuple')

    def _get_lookupds(self, lookupd_http_addresses):
        if isinstance(lookupd_http_addresses, basestring):
            return [lookupd_http_addresses]

        elif isinstance(lookupd_http_addresses, (list, tuple)):
            lookupd_http_addresses = list(lookupd_http_addresses)
            return lookupd_http_addresses

        msg = 'lookupd_http_addresses must be a list, set or tuple'
        raise TypeError(msg)

    def start(self, block=True):
        """Start discovering and listing to connections."""
        if self.state != INIT:
            self.logger.warn('%s all ready started' % self.name)
            if block:
                self.join()
            return

        if not any(self.on_message.receivers_for(blinker.ANY)):
            raise RuntimeError('no receivers connected to on_message')

        self.logger.debug('starting %s...' % self.name)
        self.state = RUNNING
        self.query_nsqd()

        if self.lookupds:
            self.query_lookupd()
            self.workers.append(gevent.spawn(self._poll_lookupd))

        self.workers.append(gevent.spawn(self._poll_ready))

        for _ in xrange(self.max_concurrency):
            self.workers.append(gevent.spawn(self._run))

        if block:
            self.join()

    def close(self):
        """Immediately close all connections and stop workers."""
        if not self.is_running:
            return

        self.state = CLOSED

        self.logger.debug('closing %d worker(s)' % len(self.workers))
        gevent.killall(self.workers, block=False)

        self.logger.debug('closing %d connection(s)' % len(self.conns))
        for conn in self.conns:
            conn.close_stream()

        self.on_close.send(self)

    def join(self, timeout=None, raise_error=False):
        """Block until all connections have closed and workers stopped."""
        gevent.joinall(self.workers, timeout, raise_error)
        gevent.joinall(self.conn_workers.values(), timeout, raise_error)

    @property
    def is_running(self):
        """Check if reader is currently running."""
        return self.state in (RUNNING, BACKOFF, THROTTLED)

    @property
    def is_starved(self):
        """Evaluate whether any of the connections are starved.

        This property should be used by message handlers to reliably identify
        when to process a batch of messages.
        """
        return any(conn.is_starved for conn in self.conns)

    @property
    def connection_max_in_flight(self):
        return max(1, self.max_in_flight / max(1, len(self.conns)))

    @property
    def total_ready_count(self):
        return sum(c.ready_count for c in self.conns)

    @property
    def total_in_flight(self):
        return sum(c.in_flight for c in self.conns)

    @property
    def total_in_flight_or_ready(self):
        return self.total_in_flight + self.total_ready_count

    def send_ready(self, conn, count):
        if self.state in (BACKOFF, CLOSED):
            return

        if self.state == THROTTLED and self.total_in_flight_or_ready:
            return

        total = self.total_ready_count - conn.ready_count + count
        if total > self.max_in_flight:
            if not (conn.ready_count or conn.in_flight):
                self.logger.debug('[%s] sending later' % conn)
                gevent.spawn_later(5, self.send_ready, conn, count)
            return

        self.logger.debug('[%s] sending RDY %d' % (conn, count))
        conn.ready(count)

    def query_nsqd(self):
        self.logger.debug('querying nsqd...')
        for address in self.nsqd_tcp_addresses:
            address, port = address.split(':')
            conn = Nsqd(address, int(port), **self.conn_kwargs)
            self.connect_to_nsqd(conn)

    def query_lookupd(self):
        self.logger.debug('querying lookupd...')
        lookupd = self.iterlookupds.next()

        try:
            producers = lookupd.lookup(self.topic)['producers']
            self.logger.debug('found %d producers' % len(producers))

        except Exception as error:
            msg = 'Failed to lookup %s on %s (%s)'
            self.logger.warn(msg % (self.topic, lookupd.address, error))
            return

        for producer in producers:
            conn = Nsqd(
                producer.get('broadcast_address') or producer['address'],
                producer['tcp_port'],
                producer['http_port'],
                **self.conn_kwargs
            )
            self.connect_to_nsqd(conn)

    def create_backoff(self):
        return BackoffTimer(max_interval=self.max_backoff_duration)

    def start_backoff(self):
        self.state = BACKOFF

        for conn in self.conns:
            conn.ready(0)

        interval = self.backoff.get_interval()
        self.logger.info('backing off for %s seconds' % interval)
        gevent.sleep(interval)

        self.state = THROTTLED
        if not self.conns:
            return

        conn = self.random_connection()
        self.logger.info('[%s] testing backoff state with RDY 1' % conn)
        self.send_ready(conn, 1)

    def complete_backoff(self):
        self.state = RUNNING
        self.logger.info('backoff complete, resuming normal operation')

        count = self.connection_max_in_flight
        for conn in self.conns:
            self.send_ready(conn, count)

    def _poll_lookupd(self):
        delay = self.lookupd_poll_interval * self.lookupd_poll_jitter
        gevent.sleep(random.random() * delay)

        while True:
            gevent.sleep(self.lookupd_poll_interval)
            self.query_lookupd()

    def _poll_ready(self):
        while True:
            gevent.sleep(5)
            if not self.need_ready_redistributed:
                continue
            self.redistribute_ready_state()

    @property
    def need_ready_redistributed(self):
        if self.state == BACKOFF:
            return False

        if self._need_ready_redistributed:
            return True

        if len(self.conns) > self.max_in_flight:
            return True

        if self.state == THROTTLED and len(self.conns) > 1:
            return True

    @need_ready_redistributed.setter
    def need_ready_redistributed(self, value):
        self._need_ready_redistributed = value

    def redistribute_ready_state(self):
        self.need_ready_redistributed = False

        # first set RDY 0 to all connections that have not received a message
        # within a configurable timeframe (low_ready_idle_timeout).
        for conn in self.conns:
            if conn.ready_count == 0:
                continue

            if (time.time() - conn.last_message) < self.low_ready_idle_timeout:
                continue

            self.logger.info('[%s] idle connection, giving up RDY count' % conn)
            conn.ready(0)

        if self.state == THROTTLED:
            max_in_flight = 1 - self.total_in_flight_or_ready
        else:
            max_in_flight = self.max_in_flight - self.total_ready_count

        if max_in_flight <= 0:
            return

        # randomly walk the list of possible connections and send RDY 1 (up to
        # our calculate "max_in_flight").  We only need to send RDY 1 because in
        # both cases described above your per connection RDY count would never
        # be higher.
        #
        # We also don't attempt to avoid the connections who previously might
        # have had RDY 1 because it would be overly complicated and not actually
        # worth it (ie. given enough redistribution rounds it doesn't matter).
        conns = list(self.conns)
        conns = random.sample(conns, min(max_in_flight, len(self.conns)))

        for conn in conns:
            self.logger.info('[%s] redistributing RDY' % conn)
            self.send_ready(conn, 1)

    def random_ready_conn(self, conn):
        # if all connections aren't getting RDY
        # occsionally randomize which connection gets RDY
        if len(self.conns) <= self.max_in_flight:
            return conn

        if (time.time() - self.last_random_ready) < 30:
            return conn

        self.last_random_ready = time.time()
        return random.choice([c for c in self.conns if not c.ready_count])

    def update_ready(self, conn):
        if self.state in (BACKOFF, THROTTLED):
            return

        conn = self.random_ready_conn(conn)
        if conn.ready_count < max(conn.last_ready * 0.25, 2):
            self.send_ready(conn, self.connection_max_in_flight)

    def random_connection(self):
        if not self.conns:
            return None
        return random.choice(list(self.conns))

    def publish(self, topic, message):
        """Publish a message to a random connection."""
        conn = self.random_connection()
        if conn is None:
            raise NSQNoConnections()

        conn.publish(topic, message)

    def connect_to_nsqd(self, conn):
        if not self.is_running:
            return

        if conn in self.conns:
            self.logger.debug('[%s] already connected' % conn)
            return

        if conn in self.pending:
            self.logger.debug('[%s] already pending' % conn)
            return

        self.logger.debug('[%s] connecting...' % conn)

        conn.on_response.connect(self.handle_response)
        conn.on_error.connect(self.handle_error)
        conn.on_finish.connect(self.handle_finish)
        conn.on_requeue.connect(self.handle_requeue)
        conn.on_auth.connect(self.handle_auth)

        if self.max_concurrency:
            conn.on_message.connect(self.queue_message)
        else:
            conn.on_message.connect(self.handle_message)

        self.pending.add(conn)

        try:
            conn.connect()
            conn.identify()

            if conn.max_ready_count < self.max_in_flight:
                msg = ' '.join([
                    '[%s] max RDY count %d < reader max in flight %d,',
                    'truncation possible'
                ])

                self.logger.warning(msg % (
                    conn,
                    conn.max_ready_count,
                    self.max_in_flight
                ))

            conn.subscribe(self.topic, self.channel)
            self.send_ready(conn, 1)

        except NSQException as error:
            self.logger.warn('[%s] connection failed (%r)' % (conn, error))
            self.handle_connection_failure(conn)
            return

        finally:
            self.pending.remove(conn)

        # Check if we've closed since we started
        if not self.is_running:
            conn.close_stream()
            return

        self.logger.info('[%s] connection successful' % conn)
        self.handle_connection_success(conn)

    def _listen(self, conn):
        try:
            conn.listen()
        except NSQException as error:
            self.logger.warning('[%s] connection lost (%r)' % (conn, error))

        self.handle_connection_failure(conn)

    def _run(self):
        for conn, message in self.queue:
            self.handle_message(conn, message)

    def queue_message(self, conn, message):
        self.logger.debug('[%s] queueing message: %s' % (conn, message.id))
        self.queue.put((conn, message))

    def handle_connection_success(self, conn):
        self.conns.add(conn)
        self.conn_workers[conn] = gevent.spawn(self._listen, conn)

        if str(conn) not in self.nsqd_tcp_addresses:
            return

        self.conn_backoffs[conn].success()

    def handle_connection_failure(self, conn):
        self.conns.discard(conn)
        self.conn_workers.pop(conn, None)
        conn.close_stream()

        if self.state == CLOSED:
            return

        if conn.ready_count:
            self.need_ready_redistributed = True

        if str(conn) not in self.nsqd_tcp_addresses:
            return

        seconds = self.conn_backoffs[conn].failure().get_interval()
        self.logger.debug('[%s] retrying in %ss' % (conn, seconds))
        gevent.spawn_later(seconds, self.connect_to_nsqd, conn)

    def handle_response(self, conn, response):
        self.logger.debug('[%s] response: %s' % (conn, response))
        self.on_response.send(self, response=response)

    def handle_error(self, conn, error):
        self.logger.debug('[%s] error: %s' % (conn, error))
        self.on_error.send(self, error=error)

    def handle_message(self, conn, message):
        self.logger.debug('[%s] got message: %s' % (conn, message.id))
        self.update_ready(conn)

        if self.max_tries and message.attempts > self.max_tries:
            msg = "giving up on message '%s' after max tries %d"
            self.logger.warning(msg % (message.id, self.max_tries))
            self.on_giving_up.send(self, message=message)
            return message.finish()

        try:
            self.on_message.send(self, message=message)
            if self.is_running and not self.async:
                message.finish()
            return

        except NSQRequeueMessage:
            pass

        except Exception as error:
            msg = '[%s] caught exception while handling message' % conn
            self.logger.exception(msg)
            self.on_exception.send(self, message=message, error=error)

        if self.is_running:
            message.requeue(self.requeue_delay)

    def handle_finish(self, conn, message_id):
        self.logger.debug('[%s] finished message: %s' % (conn, message_id))
        self.backoff.success()
        self.update_ready(conn)
        self.handle_backoff()
        self.on_finish.send(self, message_id=message_id)

    def handle_requeue(self, conn, message_id, timeout):
        msg = '[%s] requeued message: %s (%s)'
        self.logger.debug(msg % (conn, message_id, timeout))
        self.backoff.failure()
        self.update_ready(conn)
        self.handle_backoff()
        self.on_requeue.send(self, message_id=message_id, timeout=timeout)

    def handle_auth(self, conn, response):
        metadata = []
        if response.get('identity'):
            metadata.append("Identity: %r" % response['identity'])

        if response.get('permission_count'):
            metadata.append("Permissions: %d" % response['permission_count'])

        if response.get('identity_url'):
            metadata.append(response['identity_url'])

        self.logger.info('[%s] AUTH accepted %s' % (conn, ' '.join(metadata)))
        self.on_auth.send(self, conn=conn, response=response)

    def handle_backoff(self):
        if self.state in (BACKOFF, CLOSED):
            return

        if self.state == THROTTLED and self.backoff.is_reset():
            return self.complete_backoff()

        if not self.backoff.is_reset():
            return self.start_backoff()



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import re
import struct

try:
    import simplejson as json
except ImportError:
    import json  # pyflakes.ignore


__all__ = [
    'MAGIC_V2',
    'FRAME_TYPE_RESPONSE',
    'FRAME_TYPE_ERROR',
    'FRAME_TYPE_MESSAGE',
    'unpack_size',
    'unpack_response',
    'unpack_message',
    'subscribe',
    'publish',
    'ready',
    'finish',
    'requeue',
    'close',
    'nop'
]

MAGIC_V2 = '  V2'
NEWLINE = '\n'
HEARTBEAT = '_heartbeat_'

FRAME_TYPE_RESPONSE = 0
FRAME_TYPE_ERROR = 1
FRAME_TYPE_MESSAGE = 2


#
# Helpers
#
TOPIC_NAME_RE = re.compile(r'^[\.a-zA-Z0-9_-]+$')
CHANNEL_NAME_RE = re.compile(r'^[\.a-zA-Z0-9_-]+(#ephemeral)?$')


def valid_topic_name(topic):
    if not 0 < len(topic) < 33:
        return False
    return bool(TOPIC_NAME_RE.match(topic))


def valid_channel_name(channel):
    if not 0 < len(channel) < 33:
        return False
    return bool(CHANNEL_NAME_RE.match(channel))


def assert_valid_topic_name(topic):
    if valid_topic_name(topic):
        return
    raise ValueError('invalid topic name')


def assert_valid_channel_name(channel):
    if valid_channel_name(channel):
        return
    raise ValueError('invalid channel name')


#
# Responses
#
def unpack_size(data):
    return struct.unpack('>l', data)[0]


def unpack_response(data):
    return unpack_size(data[:4]), data[4:]


def unpack_message(data):
    timestamp = struct.unpack('>q', data[:8])[0]
    attempts = struct.unpack('>h', data[8:10])[0]
    message_id = data[10:26]
    body = data[26:]
    return timestamp, attempts, message_id, body


#
# Commands
#
def _packsize(data):
    return struct.pack('>l', len(data))


def _packbody(body):
    if body is None:
        return ''
    if not isinstance(body, str):
        raise TypeError('message body must be a byte string')
    return _packsize(body) + body


def _encode_param(data):
    if not isinstance(data, unicode):
        return data
    return data.encode('utf-8')


def _command(cmd, body, *params):
    params = tuple(_encode_param(p) for p in params)
    return ''.join((' '.join((cmd,) + params), NEWLINE, _packbody(body)))


def identify(data):
    return _command('IDENTIFY', json.dumps(data))


def auth(secret):
    return _command('AUTH', secret)


def subscribe(topic_name, channel_name):
    assert_valid_topic_name(topic_name)
    assert_valid_channel_name(channel_name)
    return _command('SUB', None, topic_name, channel_name)


def publish(topic_name, data):
    assert_valid_topic_name(topic_name)
    return _command('PUB', data, topic_name)


def multipublish(topic_name, messages):
    assert_valid_topic_name(topic_name)
    data = ''.join(_packbody(m) for m in messages)
    return _command('MPUB', _packsize(messages) + data, topic_name)


def ready(count):
    if not isinstance(count, int):
        raise TypeError('ready count must be an integer')

    if count < 0:
        raise ValueError('ready count cannot be negative')

    return _command('RDY', None, str(count))


def finish(message_id):
    return _command('FIN', None, message_id)


def requeue(message_id, timeout=0):
    if not isinstance(timeout, int):
        raise TypeError('requeue timeout must be an integer')
    return _command('REQ', None, message_id, str(timeout))


def touch(message_id):
    return _command('TOUCH', None, message_id)


def close():
    return _command('CLS', None)


def nop():
    return _command('NOP', None)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import random


class BackoffTimer(object):
    def __init__(self, ratio=1, max_interval=None, min_interval=None):
        self.c = 0
        self.ratio = ratio

        self.max_interval = max_interval
        self.min_interval = min_interval

    def is_reset(self):
        return self.c == 0

    def reset(self):
        self.c = 0
        return self

    def success(self):
        self.c = max(self.c - 1, 0)
        return self

    def failure(self):
        self.c += 1
        return self

    def get_interval(self):
        k = pow(2, self.c) - 1
        interval = random.random() * k * self.ratio

        if self.max_interval is not None:
            interval = min(interval, self.max_interval)

        if self.min_interval is not None:
            interval = max(interval, self.min_interval)

        return interval



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import socket


class NSQException(Exception):
    pass


class NSQRequeueMessage(NSQException):
    pass


class NSQNoConnections(NSQException):
    pass


class NSQHttpError(NSQException):
    pass


class NSQSocketError(socket.error, NSQException):
    pass


class NSQFrameError(NSQException):
    pass


class NSQErrorCode(NSQException):
    fatal = True


class NSQInvalid(NSQErrorCode):
    """E_INVALID"""
    pass


class NSQBadBody(NSQErrorCode):
    """E_BAD_BODY"""
    pass


class NSQBadTopic(NSQErrorCode):
    """E_BAD_TOPIC"""
    pass


class NSQBadChannel(NSQErrorCode):
    """E_BAD_CHANNEL"""
    pass


class NSQBadMessage(NSQErrorCode):
    """E_BAD_MESSAGE"""
    pass


class NSQPutFailed(NSQErrorCode):
    """E_PUT_FAILED"""
    pass


class NSQPubFailed(NSQErrorCode):
    """E_PUB_FAILED"""


class NSQMPubFailed(NSQErrorCode):
    """E_MPUB_FAILED"""


class NSQAuthDisabled(NSQErrorCode):
    """E_AUTH_DISABLED"""


class NSQAuthFailed(NSQErrorCode):
    """E_AUTH_FAILED"""


class NSQUnauthorized(NSQErrorCode):
    """E_UNAUTHORIZED"""


class NSQFinishFailed(NSQErrorCode):
    """E_FIN_FAILED"""
    fatal = False


class NSQRequeueFailed(NSQErrorCode):
    """E_REQ_FAILED"""
    fatal = False


class NSQTouchFailed(NSQErrorCode):
    """E_TOUCH_FAILED"""
    fatal = False


ERROR_CODES = {
    'E_INVALID': NSQInvalid,
    'E_BAD_BODY': NSQBadBody,
    'E_BAD_TOPIC': NSQBadTopic,
    'E_BAD_CHANNEL': NSQBadChannel,
    'E_BAD_MESSAGE': NSQBadMessage,
    'E_PUT_FAILED': NSQPutFailed,
    'E_PUB_FAILED': NSQPubFailed,
    'E_MPUB_FAILED': NSQMPubFailed,
    'E_FINISH_FAILED': NSQFinishFailed,
    'E_AUTH_DISABLED': NSQAuthDisabled,
    'E_AUTH_FAILED': NSQAuthFailed,
    'E_UNAUTHORIZED': NSQUnauthorized,
    'E_FIN_FAILED': NSQFinishFailed,
    'E_REQUEUE_FAILED': NSQRequeueFailed,
    'E_REQ_FAILED': NSQRequeueFailed,
    'E_TOUCH_FAILED': NSQTouchFailed
}


def make_error(error_code):
    parts = error_code.split(None, 1)
    return ERROR_CODES.get(parts[0], NSQErrorCode)(parts[-1])



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from .httpclient import HTTPClient
from . import protocol as nsq


class Lookupd(HTTPClient):
    """Low level client for nsqlookupd.

    :param address: nsqlookupd http address (default: http://localhost:4161/)
    """
    def __init__(self, address='http://localhost:4161/'):
        self.address = self.base_url = address

    def lookup(self, topic):
        """Returns producers for a topic."""
        nsq.assert_valid_topic_name(topic)
        return self.http_get('/lookup', fields={'topic': topic})

    def topics(self):
        """Returns all known topics."""
        return self.http_get('/topics')

    def channels(self, topic):
        """Returns all known channels of a topic."""
        nsq.assert_valid_topic_name(topic)
        return self.http_get('/channels', fields={'topic': topic})

    def nodes(self):
        """Returns all known nsqd."""
        return self.http_get('/nodes')

    def delete_topic(self, topic):
        """Deletes an existing topic."""
        nsq.assert_valid_topic_name(topic)
        return self.http_post('/delete_topic', fields={'topic': topic})

    def delete_channel(self, topic, channel):
        """Deletes an existing channel of an existing topic."""
        nsq.assert_valid_topic_name(topic)
        nsq.assert_valid_channel_name(channel)
        return self.http_post(
            url='/delete_channel',
            fields={'topic': topic, 'channel': channel},
        )

    def tombstone_topic_producer(self, topic, node):
        """Tombstones a specific producer of an existing topic."""
        nsq.assert_valid_topic_name(topic)
        return self.http_post(
            url='/tombstone_topic_producer',
            fields={'topic': topic, 'node': node},
        )

    def ping(self):
        """Monitoring endpoint.

        :returns: should return `"OK"`, otherwise raises an exception.
        """
        return self.http_get('/ping')

    def info(self):
        """Returns version information."""
        return self.http_get('/info')



import os
import random
import time
import shutil
import subprocess
import tempfile
import os.path
import urllib3


def with_all(head, *tail):
    def decorator(fn, *args):
        with head as arg:
            args = args + (arg,)
            if not tail:
                return fn(*args)
            return with_all(*tail)(fn, *args)
    return decorator


class BaseIntegrationServer(object):
    http = urllib3.PoolManager()

    def __init__(self, address=None, tcp_port=None, http_port=None):
        if address is None:
            address = '127.0.0.1'

        if tcp_port is None:
            tcp_port = random.randint(10000, 65535)

        if http_port is None:
            http_port = tcp_port + 1

        self.address = address
        self.tcp_port = tcp_port
        self.http_port = http_port
        self.data_path = tempfile.mkdtemp()

    @property
    def version(self):
        version = os.environ.get('NSQ_VERSION', '0.2.28')
        return tuple([int(v) for v in version.split('.')])

    @property
    def cmd(self):
        raise NotImplementedError('cmd not implemented')

    @property
    def tcp_address(self):
        return '%s:%d' % (self.address, self.tcp_port)

    @property
    def http_address(self):
        return '%s:%d' % (self.address, self.http_port)

    def is_running(self):
        try:
            url = 'http://%s/ping' % self.http_address
            return self.http.request('GET', url).data == 'OK'
        except urllib3.exceptions.HTTPError:
            return False

    def wait(self):
        for attempt in xrange(12):
            if self.is_running():
                return
            time.sleep(0.01 * pow(2, attempt))
        raise RuntimeError('unable to start: %r' % ' '.join(self.cmd))

    def __enter__(self):
        self.subp = subprocess.Popen(self.cmd)
        self.wait()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.subp.terminate()
        self.subp.wait()
        shutil.rmtree(self.data_path)


class NsqdIntegrationServer(BaseIntegrationServer):
    tls_cert = os.path.join(os.path.dirname(__file__), 'cert.pem')
    tls_key = os.path.join(os.path.dirname(__file__), 'key.pem')

    def __init__(self, lookupd=None, **kwargs):
        self.lookupd = lookupd
        super(NsqdIntegrationServer, self).__init__(**kwargs)

    @property
    def https_port(self):
        return self.http_port + 1

    @property
    def https_address(self):
        return '%s:%d' % (self.address, self.https_port)

    @property
    def cmd(self):
        cmd = [
            'nsqd',
            '--tcp-address', self.tcp_address,
            '--http-address', self.http_address,
            '--data-path', self.data_path,
            '--tls-cert', self.tls_cert,
            '--tls-key', self.tls_key,
        ]

        if self.version >= (0, 2, 28):
            cmd.extend(['--https-address', self.https_address])

        if self.lookupd:
            cmd.extend(['--lookupd-tcp-address', self.lookupd])

        return cmd


class LookupdIntegrationServer(BaseIntegrationServer):
    @property
    def cmd(self):
        return [
            'nsqlookupd',
            '--tcp-address', self.tcp_address,
            '--http-address', self.http_address,
        ]



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import blinker
import time
from gevent import socket

try:
    import simplejson as json
except ImportError:
    import json  # pyflakes.ignore

from . import protocol as nsq
from . import errors

from .message import Message
from .httpclient import HTTPClient
from .states import INIT, CONNECTED, DISCONNECTED
from .stream import Stream
from .decorators import cached_property
from .version import __version__

HOSTNAME = socket.gethostname()
SHORTNAME = HOSTNAME.split('.')[0]
USERAGENT = 'gnsq/%s' % __version__


class Nsqd(HTTPClient):
    """Low level object representing a TCP or HTTP connection to nsqd.

    :param address: the host or ip address of the nsqd

    :param tcp_port: the nsqd tcp port to connect to

    :param http_port: the nsqd http port to connect to

    :param timeout: the timeout for read/write operations (in seconds)

    :param client_id: an identifier used to disambiguate this client (defaults
        to the first part of the hostname)

    :param hostname: the hostname where the client is deployed (defaults to the
        clients hostname)

    :param heartbeat_interval: the amount of time in seconds to negotiate with
        the connected producers to send heartbeats (requires nsqd 0.2.19+)

    :param output_buffer_size: size of the buffer (in bytes) used by nsqd for
        buffering writes to this connection

    :param output_buffer_timeout: timeout (in ms) used by nsqd before flushing
        buffered writes (set to 0 to disable). Warning: configuring clients with
        an extremely low (< 25ms) output_buffer_timeout has a significant effect
        on nsqd CPU usage (particularly with > 50 clients connected).

    :param tls_v1: enable TLS v1 encryption (requires nsqd 0.2.22+)

    :param tls_options: dictionary of options to pass to `ssl.wrap_socket()
        <http://docs.python.org/2/library/ssl.html#ssl.wrap_socket>`_

    :param snappy: enable Snappy stream compression (requires nsqd 0.2.23+)

    :param deflate: enable deflate stream compression (requires nsqd 0.2.23+)

    :param deflate_level: configure the deflate compression level for this
        connection (requires nsqd 0.2.23+)

    :param sample_rate: take only a sample of the messages being sent to the
        client. Not setting this or setting it to 0 will ensure you get all the
        messages destined for the client. Sample rate can be greater than 0 or
        less than 100 and the client will receive that percentage of the message
        traffic. (requires nsqd 0.2.25+)

    :param auth_secret: a string passed when using nsq auth (requires
        nsqd 0.2.29+)

    :param user_agent: a string identifying the agent for this client in the
        spirit of HTTP (default: ``<client_library_name>/<version>``) (requires
        nsqd 0.2.25+)
    """
    def __init__(
        self,
        address='127.0.0.1',
        tcp_port=4150,
        http_port=4151,
        timeout=60.0,
        client_id=None,
        hostname=None,
        heartbeat_interval=30,
        output_buffer_size=16 * 1024,
        output_buffer_timeout=250,
        tls_v1=False,
        tls_options=None,
        snappy=False,
        deflate=False,
        deflate_level=6,
        sample_rate=0,
        auth_secret=None,
        user_agent=USERAGENT,
    ):
        self.address = address
        self.tcp_port = tcp_port
        self.http_port = http_port
        self.timeout = timeout

        self.client_id = client_id or SHORTNAME
        self.hostname = hostname or HOSTNAME
        self.heartbeat_interval = 1000 * heartbeat_interval
        self.output_buffer_size = output_buffer_size
        self.output_buffer_timeout = output_buffer_timeout
        self.tls_v1 = tls_v1
        self.tls_options = tls_options
        self.snappy = snappy
        self.deflate = deflate
        self.deflate_level = deflate_level
        self.sample_rate = sample_rate
        self.auth_secret = auth_secret
        self.user_agent = user_agent

        self.state = INIT
        self.last_response = time.time()
        self.last_message = time.time()
        self.last_ready = 0
        self.ready_count = 0
        self.in_flight = 0
        self.max_ready_count = 2500

        self._frame_handlers = {
            nsq.FRAME_TYPE_RESPONSE: self.handle_response,
            nsq.FRAME_TYPE_ERROR: self.handle_error,
            nsq.FRAME_TYPE_MESSAGE: self.handle_message
        }

    @cached_property
    def on_message(self):
        """Emitted when a message frame is received.

        The signal sender is the connection and the `message` is sent as an
        argument.
        """
        return blinker.Signal(doc='Emitted when a message frame is received.')

    @cached_property
    def on_response(self):
        """Emitted when a response frame is received.

        The signal sender is the connection and the `response` is sent as an
        argument.
        """
        return blinker.Signal(doc='Emitted when a response frame is received.')

    @cached_property
    def on_error(self):
        """Emitted when an error frame is received.

        The signal sender is the connection and the `error` is sent as an
        argument.
        """
        return blinker.Signal(doc='Emitted when a error frame is received.')

    @cached_property
    def on_finish(self):
        """Emitted after :meth:`finish`.

        Sent after a message owned by this connection is successfully finished.
        The signal sender is the connection and the `message_id` is sent as an
        argument.
        """
        return blinker.Signal(doc='Emitted after the a message is finished.')

    @cached_property
    def on_requeue(self):
        """Emitted after :meth:`requeue`.

        Sent after a message owned by this connection is requeued. The signal
        sender is the connection and the `message_id` and `timeout` are sent as
        arguments.
        """
        return blinker.Signal(doc='Emitted after the a message is requeued.')

    @cached_property
    def on_auth(self):
        """Emitted after the connection is successfully authenticated.

        The signal sender is the connection and the parsed `response` is sent as
        arguments.
        """
        return blinker.Signal(
            doc='Emitted after the connection is successfully authenticated.'
        )

    @cached_property
    def on_close(self):
        """Emitted after :meth:`close_stream`.

        Sent after the connection socket has closed. The signal sender is the
        connection.
        """
        return blinker.Signal(doc='Emitted after the connection is closed.')

    @property
    def is_connected(self):
        """Check if the client is currently connected."""
        return self.state == CONNECTED

    @property
    def is_starved(self):
        """Evaluate whether the connection is starved.

        This property should be used by message handlers to reliably identify
        when to process a batch of messages.
        """
        return self.in_flight >= max(self.last_ready * 0.85, 1)

    def connect(self):
        """Initialize connection to the nsqd."""
        if self.state not in (INIT, DISCONNECTED):
            return

        stream = Stream(self.address, self.tcp_port, self.timeout)
        stream.connect()

        self.stream = stream
        self.state = CONNECTED
        self.send(nsq.MAGIC_V2)

    def close_stream(self):
        """Close the underlying socket."""
        if not self.is_connected:
            return

        self.stream.close()
        self.state = DISCONNECTED
        self.on_close.send(self)

    def send(self, data, async=False):
        try:
            return self.stream.send(data, async)
        except Exception:
            self.close_stream()
            raise

    def _read_response(self):
        try:
            size = nsq.unpack_size(self.stream.read(4))
            return self.stream.read(size)
        except Exception:
            self.close_stream()
            raise

    def read_response(self):
        """Read an individual response from nsqd.

        :returns: tuple of the frame type and the processed data.
        """

        response = self._read_response()
        frame, data = nsq.unpack_response(response)
        self.last_response = time.time()

        if frame not in self._frame_handlers:
            raise errors.NSQFrameError('unknown frame %d' % frame)

        frame_handler = self._frame_handlers[frame]
        processed_data = frame_handler(data)

        return frame, processed_data

    def handle_response(self, data):
        if data == nsq.HEARTBEAT:
            self.nop()

        self.on_response.send(self, response=data)
        return data

    def handle_error(self, data):
        error = errors.make_error(data)
        self.on_error.send(self, error=error)

        if error.fatal:
            self.close_stream()

        return error

    def handle_message(self, data):
        self.last_message = time.time()
        self.ready_count -= 1
        self.in_flight += 1

        message = Message(*nsq.unpack_message(data))
        message.on_finish.connect(self.handle_finish)
        message.on_requeue.connect(self.handle_requeue)
        message.on_touch.connect(self.handle_touch)

        self.on_message.send(self, message=message)
        return message

    def handle_finish(self, message):
        self.finish(message.id)

    def handle_requeue(self, message, timeout):
        self.requeue(message.id, timeout)

    def handle_touch(self, message):
        self.touch(message.id)

    def finish_inflight(self):
        self.in_flight -= 1

    def listen(self):
        """Listen to incoming responses until the connection closes."""
        while self.is_connected:
            self.read_response()

    def check_ok(self, expected='OK'):
        frame, data = self.read_response()
        if frame == nsq.FRAME_TYPE_ERROR:
            raise data

        if frame != nsq.FRAME_TYPE_RESPONSE:
            raise errors.NSQException('expected response frame')

        if data != expected:
            raise errors.NSQException('unexpected response %r' % data)

    def upgrade_to_tls(self):
        self.stream.upgrade_to_tls(**self.tls_options)
        self.check_ok()

    def upgrade_to_snappy(self):
        self.stream.upgrade_to_snappy()
        self.check_ok()

    def upgrade_to_defalte(self):
        self.stream.upgrade_to_defalte(self.deflate_level)
        self.check_ok()

    def identify(self):
        """Update client metadata on the server and negotiate features.

        :returns: nsqd response data if there was feature negotiation,
            otherwise `None`
        """
        self.send(nsq.identify({
            # nsqd <0.2.28
            'short_id': self.client_id,
            'long_id': self.hostname,

            # nsqd 0.2.28+
            'client_id': self.client_id,
            'hostname': self.hostname,

            # nsqd 0.2.19+
            'feature_negotiation': True,
            'heartbeat_interval': self.heartbeat_interval,

            # nsqd 0.2.21+
            'output_buffer_size': self.output_buffer_size,
            'output_buffer_timeout': self.output_buffer_timeout,

            # nsqd 0.2.22+
            'tls_v1': self.tls_v1,

            # nsqd 0.2.23+
            'snappy': self.snappy,
            'deflate': self.deflate,
            'deflate_level': self.deflate_level,

            # nsqd nsqd 0.2.25+
            'sample_rate': self.sample_rate,
            'user_agent': self.user_agent,
        }))

        frame, data = self.read_response()

        if frame == nsq.FRAME_TYPE_ERROR:
            raise data

        if data == 'OK':
            return

        try:
            data = json.loads(data)

        except ValueError:
            self.close_stream()
            msg = 'failed to parse IDENTIFY response JSON from nsqd: %r'
            raise errors.NSQException(msg % data)

        self.max_ready_count = data.get('max_rdy_count', self.max_ready_count)

        if self.tls_v1 and data.get('tls_v1'):
            self.upgrade_to_tls()

        if self.snappy and data.get('snappy'):
            self.upgrade_to_snappy()

        elif self.deflate and data.get('deflate'):
            self.deflate_level = data.get('deflate_level', self.deflate_level)
            self.upgrade_to_defalte()

        if self.auth_secret and data.get('auth_required'):
            self.auth()

        return data

    def auth(self):
        """Send authorization secret to nsqd."""
        self.send(nsq.auth(self.auth_secret))
        frame, data = self.read_response()

        if frame == nsq.FRAME_TYPE_ERROR:
            raise data

        try:
            response = json.loads(data)
        except ValueError:
            self.close_stream()
            msg = 'failed to parse AUTH response JSON from nsqd: %r'
            raise errors.NSQException(msg % data)

        self.on_auth.send(self, response=response)
        return response

    def subscribe(self, topic, channel):
        """Subscribe to a nsq `topic` and `channel`."""
        self.send(nsq.subscribe(topic, channel))

    def publish_tcp(self, topic, data):
        """Publish a message to the given topic over tcp."""
        self.send(nsq.publish(topic, data))

    def multipublish_tcp(self, topic, messages):
        """Publish an iterable of messages to the given topic over tcp."""
        self.send(nsq.multipublish(topic, messages))

    def ready(self, count):
        """Indicate you are ready to receive `count` messages."""
        self.last_ready = count
        self.ready_count = count
        self.send(nsq.ready(count))

    def finish(self, message_id):
        """Finish a message (indicate successful processing)."""
        self.send(nsq.finish(message_id))
        self.finish_inflight()
        self.on_finish.send(self, message_id=message_id)

    def requeue(self, message_id, timeout=0):
        """Re-queue a message (indicate failure to process)."""
        self.send(nsq.requeue(message_id, timeout))
        self.finish_inflight()
        self.on_requeue.send(self, message_id=message_id, timeout=timeout)

    def touch(self, message_id):
        """Reset the timeout for an in-flight message."""
        self.send(nsq.touch(message_id))

    def close(self):
        """Indicate no more messages should be sent."""
        self.send(nsq.close())

    def nop(self):
        """Send no-op to nsqd. Used to keep connection alive."""
        self.send(nsq.nop())

    @property
    def base_url(self):
        return 'http://%s:%s/' % (self.address, self.http_port)

    def publish_http(self, topic, data):
        """Publish a message to the given topic over http."""
        nsq.assert_valid_topic_name(topic)
        return self.http_post('/put', fields={'topic': topic}, body=data)

    def _validate_http_mpub(self, message):
        if '\n' not in message:
            return message

        error = 'newlines are not allowed in http multipublish'
        raise errors.NSQException(error)

    def multipublish_http(self, topic, messages):
        """Publish an iterable of messages to the given topic over http."""
        nsq.assert_valid_topic_name(topic)
        return self.http_post(
            url='/mput',
            fields={'topic': topic},
            body='\n'.join(self._validate_http_mpub(m) for m in messages)
        )

    def create_topic(self, topic):
        """Create a topic."""
        nsq.assert_valid_topic_name(topic)
        return self.http_post('/create_topic', fields={'topic': topic})

    def delete_topic(self, topic):
        """Delete a topic."""
        nsq.assert_valid_topic_name(topic)
        return self.http_post('/delete_topic', fields={'topic': topic})

    def create_channel(self, topic, channel):
        """Create a channel for an existing topic."""
        nsq.assert_valid_topic_name(topic)
        nsq.assert_valid_channel_name(channel)
        return self.http_post(
            url='/create_channel',
            fields={'topic': topic, 'channel': channel},
        )

    def delete_channel(self, topic, channel):
        """Delete an existing channel for an existing topic."""
        nsq.assert_valid_topic_name(topic)
        nsq.assert_valid_channel_name(channel)
        return self.http_post(
            url='/delete_channel',
            fields={'topic': topic, 'channel': channel},
        )

    def empty_topic(self, topic):
        """Empty all the queued messages for an existing topic."""
        nsq.assert_valid_topic_name(topic)
        return self.http_post('/empty_topic', fields={'topic': topic})

    def empty_channel(self, topic, channel):
        """Empty all the queued messages for an existing channel."""
        nsq.assert_valid_topic_name(topic)
        nsq.assert_valid_channel_name(channel)
        return self.http_post(
            url='/empty_channel',
            fields={'topic': topic, 'channel': channel},
        )

    def pause_channel(self, topic, channel):
        """Pause message flow to all channels on an existing topic.

        Messages will queue at topic.
        """
        nsq.assert_valid_topic_name(topic)
        nsq.assert_valid_channel_name(channel)
        return self.http_post(
            url='/pause_channel',
            fields={'topic': topic, 'channel': channel},
        )

    def unpause_channel(self, topic, channel):
        """Resume message flow to channels of an existing, paused, topic."""
        nsq.assert_valid_topic_name(topic)
        nsq.assert_valid_channel_name(channel)
        return self.http_post(
            url='/unpause_channel',
            fields={'topic': topic, 'channel': channel},
        )

    def stats(self):
        """Return internal instrumented statistics."""
        return self.http_get('/stats', fields={'format': 'json'})

    def ping(self):
        """Monitoring endpoint.

        :returns: should return `"OK"`, otherwise raises an exception.
        """
        return self.http_get('/ping')

    def info(self):
        """Returns version information."""
        return self.http_get('/info')

    def publish(self, topic, data):
        """Publish a message.

        If connected, the message will be sent over tcp. Otherwise it will
        fall back to http.
        """
        if self.is_connected:
            return self.publish_tcp(topic, data)
        else:
            return self.publish_http(topic, data)

    def multipublish(self, topic, messages):
        """Publish an iterable of messages in one roundtrip.

        If connected, the messages will be sent over tcp. Otherwise it will
        fall back to http.
        """
        if self.is_connected:
            return self.multipublish_tcp(topic, messages)
        else:
            return self.multipublish_http(topic, messages)

    def __str__(self):
        return self.address + ':' + str(self.tcp_port)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return isinstance(other, Nsqd) and str(self) == str(other)

    def __cmp__(self, other):
        return hash(self) - hash(other)



import struct
import pytest

try:
    import simplejson as json
except ImportError:
    import json  # pyflakes.ignore

from gnsq import protocol as nsq


def pytest_generate_tests(metafunc):
    identify_dict_ascii = {'a': 1, 'b': 2}
    identify_dict_unicode = {'c': u'w\xc3\xa5\xe2\x80\xa0'}
    identify_body_ascii = json.dumps(identify_dict_ascii)
    identify_body_unicode = json.dumps(identify_dict_unicode)

    msgs = ['asdf', 'ghjk', 'abcd']
    mpub_body = struct.pack('>l', len(msgs)) + ''.join(struct.pack('>l', len(m)) + m for m in msgs)
    if metafunc.function == test_command:
        for cmd_method, kwargs, result in [
                ('identify',
                    {'data': identify_dict_ascii},
                    'IDENTIFY\n' + struct.pack('>l', len(identify_body_ascii)) +
                    identify_body_ascii),
                ('identify',
                    {'data': identify_dict_unicode},
                    'IDENTIFY\n' + struct.pack('>l', len(identify_body_unicode)) +
                    identify_body_unicode),
                ('auth',
                    {'secret': 'secret'},
                    'AUTH\n' + struct.pack('>l', 6) + 'secret'),
                ('subscribe',
                    {'topic_name': 'test_topic', 'channel_name': 'test_channel'},
                    'SUB test_topic test_channel\n'),
                ('finish',
                    {'message_id': 'test'},
                    'FIN test\n'),
                ('finish',
                    {'message_id': u'\u2020est \xfcn\xee\xe7\xf8\u2202\xe9'},
                    'FIN \xe2\x80\xa0est \xc3\xbcn\xc3\xae\xc3\xa7\xc3\xb8\xe2\x88\x82\xc3\xa9\n'),
                ('requeue',
                    {'message_id': 'test'},
                    'REQ test 0\n'),
                ('requeue',
                    {'message_id': 'test', 'timeout': 60},
                    'REQ test 60\n'),
                ('touch',
                    {'message_id': 'test'},
                    'TOUCH test\n'),
                ('ready',
                    {'count': 100},
                    'RDY 100\n'),
                ('nop',
                    {},
                    'NOP\n'),
                ('publish',
                    {'topic_name': 'test', 'data': msgs[0]},
                    'PUB test\n' + struct.pack('>l', len(msgs[0])) + msgs[0]),
                ('multipublish',
                    {'topic_name': 'test', 'messages': msgs},
                    'MPUB test\n' + struct.pack('>l', len(mpub_body)) + mpub_body)
                ]:
            metafunc.addcall(funcargs=dict(cmd_method=cmd_method, kwargs=kwargs, result=result))


def test_command(cmd_method, kwargs, result):
    assert getattr(nsq, cmd_method)(**kwargs) == result


def test_unicode_body():
    pytest.raises(TypeError, nsq.publish, 'topic', u'unicode body')



# -*- coding: utf-8 -*-


class cached_property(object):
    """A decorator that converts a function into a lazy property."""

    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __get__(self, obj, type=None):
        if obj is None:
            return self

        if self.__name__ in obj.__dict__:
            return obj.__dict__[self.__name__]

        value = obj.__dict__[self.__name__] = self.func(obj)
        return value



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')


setup(
    name='gnsq',
    version='0.2.0',
    description='A gevent based python client for NSQ.',
    long_description=readme + '\n\n' + history,
    author='Trevor Olson',
    author_email='trevor@heytrevor.com',
    url='https://github.com/wtolson/gnsq',
    packages=[
        'gnsq',
        'gnsq.stream',
    ],
    package_dir={'gnsq': 'gnsq'},
    include_package_data=True,
    install_requires=[
        'gevent',
        'blinker',
        'urllib3',
    ],
    extras_require={
        'snappy': ['python-snappy'],
    },
    license="BSD",
    zip_safe=False,
    keywords='gnsq',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ]
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from errno import EWOULDBLOCK
from gnsq.errors import NSQSocketError


class CompressionSocket(object):
    def __init__(self, socket):
        self._socket = socket
        self._bootstrapped = None

    def __getattr__(self, name):
        return getattr(self._socket, name)

    def bootstrap(self, data):
        if not data:
            return
        self._bootstrapped = self.decompress(data)

    def recv(self, size):
        if self._bootstrapped:
            data = self._bootstrapped
            self._bootstrapped = None
            return data

        chunk = self._socket.recv(size)
        if chunk:
            uncompressed = self.decompress(chunk)

        if not uncompressed:
            raise NSQSocketError(EWOULDBLOCK, 'Operation would block')

        return uncompressed

    def send(self, data):
        self._socket.send(self.compress(data))
